@@ -25,6 +25,7 @@
 #include <vector>
 #include <string>
 #include <memory>
+#include <set>
 
 
 class ImageGrid
@@ -89,18 +90,18 @@ public:
                                                                    uint32_t output_height,
                                                                    uint16_t tile_rows,
                                                                    uint16_t tile_columns,
-                                                                   const struct heif_encoding_options* encoding_options);
+                                                                   const heif_encoding_options* encoding_options);
 
   Error add_image_tile(uint32_t tile_x, uint32_t tile_y,
                        const std::shared_ptr<HeifPixelImage>& image,
-                       struct heif_encoder* encoder);
+                       heif_encoder* encoder);
 
   static Result<std::shared_ptr<ImageItem_Grid>> add_and_encode_full_grid(HeifContext* ctx,
                                                                           const std::vector<std::shared_ptr<HeifPixelImage>>& tiles,
                                                                           uint16_t rows,
                                                                           uint16_t columns,
-                                                                          struct heif_encoder* encoder,
-                                                                          const struct heif_encoding_options& options);
+                                                                          heif_encoder* encoder,
+                                                                          const heif_encoding_options& options);
 
 
   // TODO: nclx depends on contained format
@@ -108,7 +109,7 @@ public:
 
   // heif_compression_format get_compression_format() const override { return heif_compression_HEVC; }
 
-  Error on_load_file() override;
+  Error initialize_decoder() override;
 
   int get_luma_bits_per_pixel() const override;
 
@@ -121,16 +122,17 @@ public:
   const heif_encoding_options* get_encoding_options() const { return m_encoding_options; }
 
   Result<Encoder::CodedImageData> encode(const std::shared_ptr<HeifPixelImage>& image,
-                                         struct heif_encoder* encoder,
-                                         const struct heif_encoding_options& options,
-                                         enum heif_image_input_class input_class) override
+                                         heif_encoder* encoder,
+                                         const heif_encoding_options& options,
+                                         heif_image_input_class input_class) override
   {
     return Error{heif_error_Unsupported_feature,
                  heif_suberror_Unspecified, "Cannot encode image to 'grid'"};
   }
 
-  Result<std::shared_ptr<HeifPixelImage>> decode_compressed_image(const struct heif_decoding_options& options,
-                                                                  bool decode_tile_only, uint32_t tile_x0, uint32_t tile_y0) const override;
+  Result<std::shared_ptr<HeifPixelImage>> decode_compressed_image(const heif_decoding_options& options,
+                                                                  bool decode_tile_only, uint32_t tile_x0, uint32_t tile_y0,
+                                                                  std::set<heif_item_id> processed_ids) const override;
 
   heif_brand2 get_compatible_brand() const override;
 
@@ -161,13 +163,15 @@ private:
 
   Error read_grid_spec();
 
-  Result<std::shared_ptr<HeifPixelImage>> decode_full_grid_image(const heif_decoding_options& options) const;
+  Result<std::shared_ptr<HeifPixelImage>> decode_full_grid_image(const heif_decoding_options& options, std::set<heif_item_id> processed_ids) const;
 
-  Result<std::shared_ptr<HeifPixelImage>> decode_grid_tile(const heif_decoding_options& options, uint32_t tx, uint32_t ty) const;
+  Result<std::shared_ptr<HeifPixelImage>> decode_grid_tile(const heif_decoding_options& options, uint32_t tx, uint32_t ty, std::set<heif_item_id> processed_ids) const;
 
   Error decode_and_paste_tile_image(heif_item_id tileID, uint32_t x0, uint32_t y0,
                                     std::shared_ptr<HeifPixelImage>& inout_image,
-                                    const heif_decoding_options& options, int& progress_counter) const;
+                                    const heif_decoding_options& options, int& progress_counter,
+                                    std::shared_ptr<std::vector<Error> > warnings,
+                                    std::set<heif_item_id> processed_ids) const;
 };
 
 
