@@ -19,11 +19,11 @@
  */
 
 #include "init.h"
-#include "heif.h"
+#include "libheif/heif.h"
 #include "error.h"
 #include "plugin_registry.h"
 #include "common_utils.h"
-#include "libheif/color-conversion/colorconversion.h"
+#include "color-conversion/colorconversion.h"
 
 #if ENABLE_MULTITHREADING_SUPPORT
 
@@ -39,6 +39,9 @@
 
 #endif
 
+#define STRINGIFY2(x) #x
+#define STRINGIFY(x) STRINGIFY2(x)
+
 void heif_unload_all_plugins();
 
 #if ENABLE_PLUGIN_LOADING
@@ -102,7 +105,7 @@ void load_plugins_if_not_initialized_yet()
 }
 
 
-struct heif_error heif_init(struct heif_init_params*)
+heif_error heif_init(heif_init_params*)
 {
 #if ENABLE_MULTITHREADING_SUPPORT
   std::lock_guard<std::recursive_mutex> lock(heif_init_mutex());
@@ -131,8 +134,8 @@ struct heif_error heif_init(struct heif_init_params*)
 #endif
   }
 
-  // Note: it is important that we increase the counter AFTER initialization such that 'load_plugins_if_not_initialized_yet()' can check this
-  // without having to lock the mutex.
+  // Note: it is important that we increase the counter AFTER initialization such that
+  // 'load_plugins_if_not_initialized_yet()' can check this without having to lock the mutex.
   heif_library_initialization_count++;
 
   return {heif_error_Ok, heif_suberror_Unspecified, Error::kSuccess};
@@ -146,7 +149,7 @@ void heif_deinit()
 #endif
 
   if (heif_library_initialization_count == 0) {
-    // This case should never happen (heif_deinit() is called more often then heif_init()).
+    // This case should never happen (heif_deinit() is called more often than heif_init()).
     return;
   }
 
@@ -246,6 +249,22 @@ struct heif_error heif_load_plugin(const char* filename, struct heif_plugin_info
   switch (loadedPlugin.info->type) {
     case heif_plugin_type_encoder: {
       auto* encoder_plugin = static_cast<const heif_encoder_plugin*>(plugin_info->plugin);
+      if (encoder_plugin->plugin_api_version < heif_encoder_plugin_minimum_version) {
+        return {heif_error_Plugin_loading_error,
+                heif_suberror_Unsupported_plugin_version,
+                "Encoder plugin needs to be at least version " STRINGIFY(heif_encoder_plugin_latest_version)
+        };
+      }
+
+      if (encoder_plugin->plugin_api_version >= 4 &&
+          encoder_plugin->minimum_required_libheif_version > heif_get_version_number()) {
+        return {
+          heif_error_Plugin_loading_error,
+          heif_suberror_Unsupported_plugin_version,
+          "Encoder plugin requires at least libheif version " LIBHEIF_VERSION
+        };
+      }
+
       struct heif_error err = heif_register_encoder_plugin(encoder_plugin);
       if (err.code) {
         return err;
@@ -255,6 +274,22 @@ struct heif_error heif_load_plugin(const char* filename, struct heif_plugin_info
 
     case heif_plugin_type_decoder: {
       auto* decoder_plugin = static_cast<const heif_decoder_plugin*>(plugin_info->plugin);
+      if (decoder_plugin->plugin_api_version < heif_decoder_plugin_minimum_version) {
+        return {heif_error_Plugin_loading_error,
+                heif_suberror_Unsupported_plugin_version,
+                "Decoder plugin needs to be at least version " STRINGIFY(heif_decoder_plugin_latest_version)
+        };
+      }
+
+      if (decoder_plugin->plugin_api_version >= 4 &&
+          decoder_plugin->minimum_required_libheif_version > heif_get_version_number()) {
+        return {
+          heif_error_Plugin_loading_error,
+          heif_suberror_Unsupported_plugin_version,
+          "Decoder plugin requires at least libheif version " LIBHEIF_VERSION
+        };
+      }
+
       struct heif_error err = heif_register_decoder_plugin(decoder_plugin);
       if (err.code) {
         return err;
@@ -320,6 +355,10 @@ struct heif_error heif_load_plugins(const char* directory,
 
   int nPlugins = 0;
 
+  // Loading the plugins may return several errors, but we can only return one of them,
+  // which is remembered here.
+  heif_error err_result = heif_error_ok;
+
   for (const auto& filename : libraryFiles) {
     const struct heif_plugin_info* info = nullptr;
     auto err = heif_load_plugin(filename.c_str(), &info);
@@ -334,6 +373,10 @@ struct heif_error heif_load_plugins(const char* directory,
 
       nPlugins++;
     }
+    else {
+      // remember error
+      err_result = err;
+    }
   }
 
   if (nPlugins < output_array_size && out_plugins) {
@@ -344,7 +387,7 @@ struct heif_error heif_load_plugins(const char* directory,
     *out_nPluginsLoaded = nPlugins;
   }
 
-  return heif_error_ok;
+  return err_result;
 }
 
 #else
