@@ -26,6 +26,7 @@
 #include "error.h"
 #include "nclx.h"
 #include <libheif/heif_experimental.h>
+#include "security_limits.h"
 
 #include <vector>
 #include <memory>
@@ -33,7 +34,7 @@
 #include <set>
 #include <utility>
 #include <cassert>
-
+#include <string>
 
 heif_chroma chroma_from_subsampling(int h, int v);
 
@@ -71,13 +72,143 @@ enum heif_chroma420_sample_position {
 };
 
 
+class ImageExtraData
+{
+public:
+  virtual ~ImageExtraData();
+
+  std::vector<std::shared_ptr<Box>> generate_property_boxes() const;
+
+
+  // --- color profile
+
+  bool has_nclx_color_profile() const;
+
+  virtual void set_color_profile_nclx(const nclx_profile& profile) { m_color_profile_nclx = profile; }
+
+  nclx_profile get_color_profile_nclx() const { return m_color_profile_nclx; }
+
+  // get the stored nclx fallback or return the default nclx if none is stored
+  nclx_profile get_color_profile_nclx_with_fallback() const;
+
+  virtual void set_color_profile_icc(const std::shared_ptr<const color_profile_raw>& profile) { m_color_profile_icc = profile; }
+
+  bool has_icc_color_profile() const { return m_color_profile_icc != nullptr; }
+
+  const std::shared_ptr<const color_profile_raw>& get_color_profile_icc() const { return m_color_profile_icc; }
+
+  void set_color_profile(const std::shared_ptr<const color_profile>& profile)
+  {
+    auto icc = std::dynamic_pointer_cast<const color_profile_raw>(profile);
+    if (icc) {
+      set_color_profile_icc(icc);
+    }
+
+    auto nclx = std::dynamic_pointer_cast<const color_profile_nclx>(profile);
+    if (nclx) {
+      set_color_profile_nclx(nclx->get_nclx_color_profile());
+    }
+  }
+
+
+  // --- premultiplied alpha
+
+  bool is_premultiplied_alpha() const { return m_premultiplied_alpha; }
+
+  virtual void set_premultiplied_alpha(bool flag) { m_premultiplied_alpha = flag; }
+
+
+  // --- pixel aspect ratio
+
+  bool has_nonsquare_pixel_ratio() const { return m_PixelAspectRatio_h != m_PixelAspectRatio_v; }
+
+  void get_pixel_ratio(uint32_t* h, uint32_t* v) const
+  {
+    *h = m_PixelAspectRatio_h;
+    *v = m_PixelAspectRatio_v;
+  }
+
+  virtual void set_pixel_ratio(uint32_t h, uint32_t v)
+  {
+    m_PixelAspectRatio_h = h;
+    m_PixelAspectRatio_v = v;
+  }
+
+  // --- clli
+
+  bool has_clli() const { return m_clli.max_content_light_level != 0 || m_clli.max_pic_average_light_level != 0; }
+
+  heif_content_light_level get_clli() const { return m_clli; }
+
+  virtual void set_clli(const heif_content_light_level& clli) { m_clli = clli; }
+
+  // --- mdcv
+
+  bool has_mdcv() const { return m_mdcv.has_value(); }
+
+  heif_mastering_display_colour_volume get_mdcv() const { return *m_mdcv; }
+
+  virtual void set_mdcv(const heif_mastering_display_colour_volume& mdcv)
+  {
+    m_mdcv = mdcv;
+  }
+
+  void unset_mdcv() { m_mdcv.reset(); }
+
+  virtual Error set_tai_timestamp(const heif_tai_timestamp_packet* tai) {
+    delete m_tai_timestamp;
+
+    m_tai_timestamp = heif_tai_timestamp_packet_alloc();
+    heif_tai_timestamp_packet_copy(m_tai_timestamp, tai);
+    return Error::Ok;
+  }
+
+  const heif_tai_timestamp_packet* get_tai_timestamp() const {
+    return m_tai_timestamp;
+  }
+
+
+  virtual void set_gimi_sample_content_id(std::string id) { m_gimi_sample_content_id = id; }
+
+  bool has_gimi_sample_content_id() const { return m_gimi_sample_content_id.has_value(); }
+
+  std::string get_gimi_sample_content_id() const { assert(has_gimi_sample_content_id()); return *m_gimi_sample_content_id; }
+
+private:
+  bool m_premultiplied_alpha = false;
+  nclx_profile m_color_profile_nclx = nclx_profile::undefined();
+  std::shared_ptr<const color_profile_raw> m_color_profile_icc;
+
+  uint32_t m_PixelAspectRatio_h = 1;
+  uint32_t m_PixelAspectRatio_v = 1;
+  heif_content_light_level m_clli{};
+  std::optional<heif_mastering_display_colour_volume> m_mdcv;
+
+  heif_tai_timestamp_packet* m_tai_timestamp = nullptr;
+
+  std::optional<std::string> m_gimi_sample_content_id;
+
+protected:
+  std::shared_ptr<Box_clli> get_clli_box() const;
+
+  std::shared_ptr<Box_mdcv> get_mdcv_box() const;
+
+  std::shared_ptr<Box_pasp> get_pasp_box() const;
+
+  std::shared_ptr<Box_colr> get_colr_box_nclx() const;
+
+  std::shared_ptr<Box_colr> get_colr_box_icc() const;
+};
+
+
 class HeifPixelImage : public std::enable_shared_from_this<HeifPixelImage>,
+                       public ImageExtraData,
                        public ErrorBuffer
 {
 public:
   explicit HeifPixelImage() = default;
 
-  ~HeifPixelImage();
+  ~HeifPixelImage() override;
 
   void create(uint32_t width, uint32_t height, heif_colorspace colorspace, heif_chroma chroma);
 
@@ -94,17 +225,13 @@ public:
   // Has alpha information either as a separate channel or in the interleaved format.
   bool has_alpha() const;
 
-  bool is_premultiplied_alpha() const { return m_premultiplied_alpha; }
-
-  void set_premultiplied_alpha(bool flag) { m_premultiplied_alpha = flag; }
-
   uint32_t get_width() const { return m_width; }
 
   uint32_t get_height() const { return m_height; }
 
-  uint32_t get_width(enum heif_channel channel) const;
+  uint32_t get_width(heif_channel channel) const;
 
-  uint32_t get_height(enum heif_channel channel) const;
+  uint32_t get_height(heif_channel channel) const;
 
   bool has_odd_width() const { return !!(m_width & 1); }
 
@@ -114,28 +241,28 @@ public:
 
   heif_colorspace get_colorspace() const { return m_colorspace; }
 
-  std::set<enum heif_channel> get_channel_set() const;
+  std::set<heif_channel> get_channel_set() const;
 
-  uint8_t get_storage_bits_per_pixel(enum heif_channel channel) const;
+  uint8_t get_storage_bits_per_pixel(heif_channel channel) const;
 
-  uint8_t get_bits_per_pixel(enum heif_channel channel) const;
+  uint8_t get_bits_per_pixel(heif_channel channel) const;
 
   // Get the maximum bit depth of a visual channel (YCbCr or RGB).
   uint8_t get_visual_image_bits_per_pixel() const;
 
-  heif_channel_datatype get_datatype(enum heif_channel channel) const;
+  heif_channel_datatype get_datatype(heif_channel channel) const;
 
   int get_number_of_interleaved_components(heif_channel channel) const;
 
   // Note: we are using size_t as stride type since the stride is usually involved in a multiplication with the line number.
   //       For very large images (e.g. >2 GB), this can result in an integer overflow and corresponding illegal memory access.
   //       (see https://github.com/strukturag/libheif/issues/1419)
-  uint8_t* get_plane(enum heif_channel channel, size_t* out_stride) { return get_channel<uint8_t>(channel, out_stride); }
+  uint8_t* get_plane(heif_channel channel, size_t* out_stride) { return get_channel<uint8_t>(channel, out_stride); }
 
-  const uint8_t* get_plane(enum heif_channel channel, size_t* out_stride) const { return get_channel<uint8_t>(channel, out_stride); }
+  const uint8_t* get_plane(heif_channel channel, size_t* out_stride) const { return get_channel<uint8_t>(channel, out_stride); }
 
   template <typename T>
-  T* get_channel(enum heif_channel channel, size_t* out_stride)
+  T* get_channel(heif_channel channel, size_t* out_stride)
   {
     auto iter = m_planes.find(channel);
     if (iter == m_planes.end()) {
@@ -155,7 +282,7 @@ public:
   }
 
   template <typename T>
-  const T* get_channel(enum heif_channel channel, size_t* out_stride) const
+  const T* get_channel(heif_channel channel, size_t* out_stride) const
   {
     return const_cast<HeifPixelImage*>(this)->get_channel<T>(channel, out_stride);
   }
@@ -191,14 +318,6 @@ public:
   Error scale_nearest_neighbor(std::shared_ptr<HeifPixelImage>& output, uint32_t width, uint32_t height,
                                const heif_security_limits* limits) const;
 
-  void set_color_profile_nclx(const std::shared_ptr<const color_profile_nclx>& profile) { m_color_profile_nclx = profile; }
-
-  const std::shared_ptr<const color_profile_nclx>& get_color_profile_nclx() const { return m_color_profile_nclx; }
-
-  void set_color_profile_icc(const std::shared_ptr<const color_profile_raw>& profile) { m_color_profile_icc = profile; }
-
-  const std::shared_ptr<const color_profile_raw>& get_color_profile_icc() const { return m_color_profile_icc; }
-
   void forward_all_metadata_from(const std::shared_ptr<const HeifPixelImage>& src_image);
 
   void debug_dump() const;
@@ -212,43 +331,11 @@ public:
                                                              const heif_security_limits* limits) const;
 
 
-  // --- pixel aspect ratio
+  // --- sequences
 
-  bool has_nonsquare_pixel_ratio() const { return m_PixelAspectRatio_h != m_PixelAspectRatio_v; }
+  void set_sample_duration(uint32_t d) { m_sample_duration = d; }
 
-  void get_pixel_ratio(uint32_t* h, uint32_t* v) const
-  {
-    *h = m_PixelAspectRatio_h;
-    *v = m_PixelAspectRatio_v;
-  }
-
-  void set_pixel_ratio(uint32_t h, uint32_t v)
-  {
-    m_PixelAspectRatio_h = h;
-    m_PixelAspectRatio_v = v;
-  }
-
-  // --- clli
-
-  bool has_clli() const { return m_clli.max_content_light_level != 0 || m_clli.max_pic_average_light_level != 0; }
-
-  heif_content_light_level get_clli() const { return m_clli; }
-
-  void set_clli(const heif_content_light_level& clli) { m_clli = clli; }
-
-  // --- mdcv
-
-  bool has_mdcv() const { return m_mdcv_set; }
-
-  heif_mastering_display_colour_volume get_mdcv() const { return m_mdcv; }
-
-  void set_mdcv(const heif_mastering_display_colour_volume& mdcv)
-  {
-    m_mdcv = mdcv;
-    m_mdcv_set = true;
-  }
-
-  void unset_mdcv() { m_mdcv_set = false; }
+  uint32_t get_sample_duration() const { return m_sample_duration; }
 
   // --- warnings
 
@@ -262,8 +349,10 @@ private:
   struct ImagePlane
   {
     // limits=nullptr disables the limits
-    Error alloc(uint32_t width, uint32_t height, heif_channel_datatype datatype, int bit_depth, int num_interleaved_components,
-                const heif_security_limits* limits);
+    Error alloc(uint32_t width, uint32_t height, heif_channel_datatype datatype, int bit_depth,
+                int num_interleaved_components,
+                const heif_security_limits* limits,
+                MemoryHandle& memory_handle);
 
     heif_channel_datatype m_datatype = heif_channel_datatype_unsigned_integer;
     uint8_t m_bit_depth = 0;
@@ -279,7 +368,8 @@ private:
 
     void* mem = nullptr; // aligned memory start
     uint8_t* allocated_mem = nullptr; // unaligned memory we allocated
-    uint32_t stride = 0; // bytes per line
+    size_t   allocation_size = 0;
+    size_t   stride = 0; // bytes per line
 
     int get_bytes_per_pixel() const;
 
@@ -295,17 +385,11 @@ private:
   uint32_t m_height = 0;
   heif_colorspace m_colorspace = heif_colorspace_undefined;
   heif_chroma m_chroma = heif_chroma_undefined;
-  bool m_premultiplied_alpha = false;
-  std::shared_ptr<const color_profile_nclx> m_color_profile_nclx;
-  std::shared_ptr<const color_profile_raw> m_color_profile_icc;
 
   std::map<heif_channel, ImagePlane> m_planes;
+  MemoryHandle m_memory_handle;
 
-  uint32_t m_PixelAspectRatio_h = 1;
-  uint32_t m_PixelAspectRatio_v = 1;
-  heif_content_light_level m_clli{};
-  heif_mastering_display_colour_volume m_mdcv{};
-  bool m_mdcv_set = false; // replace with std::optional<> when we are on C*+17
+  uint32_t m_sample_duration = 0; // duration of a sequence frame
 
   std::vector<Error> m_warnings;
 };
