@@ -55,7 +55,7 @@ TrackOptions& TrackOptions::operator=(const TrackOptions& src)
 
 
 SampleAuxInfoHelper::SampleAuxInfoHelper(bool interleaved)
-    : m_interleaved(interleaved)
+  : m_interleaved(interleaved)
 {
   m_saiz = std::make_shared<Box_saiz>();
   m_saio = std::make_shared<Box_saio>();
@@ -71,9 +71,11 @@ void SampleAuxInfoHelper::set_aux_info_type(uint32_t aux_info_type, uint32_t aux
 Error SampleAuxInfoHelper::add_sample_info(const std::vector<uint8_t>& data)
 {
   if (data.size() > 0xFF) {
-    return {heif_error_Encoding_error,
-            heif_suberror_Unspecified,
-            "Encoded sample auxiliary information exceeds maximum size"};
+    return {
+      heif_error_Encoding_error,
+      heif_suberror_Unspecified,
+      "Encoded sample auxiliary information exceeds maximum size"
+    };
   }
 
   m_saiz->add_sample_size(static_cast<uint8_t>(data.size()));
@@ -89,9 +91,10 @@ void SampleAuxInfoHelper::add_nonpresent_sample()
 }
 
 
-void SampleAuxInfoHelper::write_interleaved(const std::shared_ptr<class HeifFile>& file)
+void SampleAuxInfoHelper::write_interleaved(const std::shared_ptr<HeifFile>& file)
 {
   if (m_interleaved && !m_data.empty()) {
+    // TODO: I think this does not work because the image data does not know that there is SAI in-between
     uint64_t pos = file->append_mdat_data(m_data);
     m_saio->add_sample_offset(pos);
 
@@ -99,7 +102,7 @@ void SampleAuxInfoHelper::write_interleaved(const std::shared_ptr<class HeifFile
   }
 }
 
-void SampleAuxInfoHelper::write_all(const std::shared_ptr<class Box>& parent, const std::shared_ptr<class HeifFile>& file)
+void SampleAuxInfoHelper::write_all(const std::shared_ptr<Box>& parent, const std::shared_ptr<HeifFile>& file)
 {
   parent->append_child_box(m_saiz);
   parent->append_child_box(m_saio);
@@ -122,7 +125,7 @@ SampleAuxInfoReader::SampleAuxInfoReader(std::shared_ptr<Box_saiz> saiz,
     uint64_t offset = saio->get_sample_offset(0);
     auto nSamples = saiz->get_num_samples();
 
-    for (uint32_t i=0;i<nSamples;i++) {
+    for (uint32_t i = 0; i < nSamples; i++) {
       m_contiguous_offsets.push_back(offset);
       offset += saiz->get_sample_size(i);
     }
@@ -141,7 +144,7 @@ heif_sample_aux_info_type SampleAuxInfoReader::get_type() const
 }
 
 
-Result<std::vector<uint8_t>> SampleAuxInfoReader::get_sample_info(const HeifFile* file, uint32_t idx)
+Result<std::vector<uint8_t> > SampleAuxInfoReader::get_sample_info(const HeifFile* file, uint32_t idx)
 {
   uint64_t offset;
   if (m_contiguous) {
@@ -163,92 +166,189 @@ Result<std::vector<uint8_t>> SampleAuxInfoReader::get_sample_info(const HeifFile
 }
 
 
-std::shared_ptr<class HeifFile> Track::get_file() const
+std::shared_ptr<HeifFile> Track::get_file() const
 {
   return m_heif_context->get_heif_file();
 }
 
 
-Track::Track(HeifContext* ctx, const std::shared_ptr<Box_trak>& trak_box)
+Track::Track(HeifContext* ctx)
 {
   m_heif_context = ctx;
+}
 
+
+Error Track::load(const std::shared_ptr<Box_trak>& trak_box)
+{
   m_trak = trak_box;
 
   auto tkhd = trak_box->get_child_box<Box_tkhd>();
   if (!tkhd) {
-    return; // TODO: error or dummy error track ?
+    return {
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Track has no 'tkhd' box."
+    };
   }
 
+  m_tkhd = tkhd;
+
   m_id = tkhd->get_track_id();
 
+  auto edts = trak_box->get_child_box<Box_edts>();
+  if (edts) {
+    m_elst = edts->get_child_box<Box_elst>();
+  }
+
   auto mdia = trak_box->get_child_box<Box_mdia>();
   if (!mdia) {
-    return;
+    return {
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Track has no 'mdia' box."
+    };
   }
 
   m_tref = trak_box->get_child_box<Box_tref>();
 
   auto hdlr = mdia->get_child_box<Box_hdlr>();
   if (!hdlr) {
-    return;
+    return {
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Track has no 'hdlr' box."
+    };
   }
 
   m_handler_type = hdlr->get_handler_type();
 
   m_minf = mdia->get_child_box<Box_minf>();
   if (!m_minf) {
-    return;
+    return {
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Track has no 'minf' box."
+    };
   }
 
   m_mdhd = mdia->get_child_box<Box_mdhd>();
   if (!m_mdhd) {
-    return;
+    return {
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Track has no 'mdhd' box."
+    };
   }
 
   auto stbl = m_minf->get_child_box<Box_stbl>();
   if (!stbl) {
-    return;
+    return {
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Track has no 'stbl' box."
+    };
   }
 
   m_stsd = stbl->get_child_box<Box_stsd>();
   if (!m_stsd) {
-    return;
+    return {
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Track has no 'stsd' box."
+    };
   }
 
   m_stsc = stbl->get_child_box<Box_stsc>();
   if (!m_stsc) {
-    return;
+    return {
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Track has no 'stsc' box."
+    };
   }
 
   m_stco = stbl->get_child_box<Box_stco>();
   if (!m_stco) {
-    return;
+    return {
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Track has no 'stco' box."
+    };
   }
 
   m_stsz = stbl->get_child_box<Box_stsz>();
   if (!m_stsz) {
-    return;
+    return {
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Track has no 'stsz' box."
+    };
   }
 
   m_stts = stbl->get_child_box<Box_stts>();
+  if (!m_stts) {
+    return {
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Track has no 'stts' box."
+    };
+  }
+
+  // --- check that number of samples in various boxes are consistent
+
+  if (m_stts->get_number_of_samples() != m_stsz->num_samples()) {
+    return {
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Number of samples in 'stts' and 'stsz' is inconsistent."
+    };
+  }
+
+  if (m_ctts && m_ctts->get_number_of_samples() != m_stsz->num_samples()) {
+    return {
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Number of samples in 'ctts' and 'stsz' is inconsistent."
+    };
+  }
+
+  if (m_stsc->get_number_of_samples() != m_stsz->num_samples()) {
+    return {
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Number of samples in 'stsc' and 'stsz' is inconsistent."
+    };
+  }
 
   const std::vector<uint32_t>& chunk_offsets = m_stco->get_offsets();
   assert(chunk_offsets.size() <= (size_t) std::numeric_limits<uint32_t>::max()); // There cannot be more than uint32_t chunks.
 
   uint32_t current_sample_idx = 0;
+  int32_t previous_sample_description_index = -1;
 
   for (size_t chunk_idx = 0; chunk_idx < chunk_offsets.size(); chunk_idx++) {
     auto* s2c = m_stsc->get_chunk(static_cast<uint32_t>(chunk_idx + 1));
     if (!s2c) {
-      return;
+      return {
+        heif_error_Invalid_input,
+        heif_suberror_Unspecified,
+        "'stco' box references a non-existing chunk."
+      };
     }
 
     Box_stsc::SampleToChunk sampleToChunk = *s2c;
 
     auto sample_description = m_stsd->get_sample_entry(sampleToChunk.sample_description_index - 1);
     if (!sample_description) {
-      return;
+      return {
+        heif_error_Invalid_input,
+        heif_suberror_Unspecified,
+        "Track references a non-existing sample description."
+      };
+    }
+
+    if (auto auxi = sample_description->get_child_box<Box_auxi>()) {
+      m_auxiliary_info_type = auxi->get_aux_track_type_urn();
     }
 
     if (m_first_taic == nullptr) {
@@ -258,20 +358,40 @@ Track::Track(HeifContext* ctx, const std::shared_ptr<Box_trak>& trak_box)
       }
     }
 
-    auto chunk = std::make_shared<Chunk>(ctx, m_id, sample_description,
+    auto chunk = std::make_shared<Chunk>(m_heif_context, m_id,
                                          current_sample_idx, sampleToChunk.samples_per_chunk,
                                          m_stco->get_offsets()[chunk_idx],
                                          m_stsz);
 
+    if (auto visualSampleDescription = std::dynamic_pointer_cast<const Box_VisualSampleEntry>(sample_description)) {
+      if (chunk_idx > 0 && (int32_t) sampleToChunk.sample_description_index == previous_sample_description_index) {
+        // reuse decoder from previous chunk if it uses the sample sample_description_index
+        chunk->set_decoder(m_chunks[chunk_idx - 1]->get_decoder());
+      }
+      else {
+        // use a new decoder
+        chunk->set_decoder(Decoder::alloc_for_sequence_sample_description_box(visualSampleDescription));
+      }
+    }
+
     m_chunks.push_back(chunk);
 
     current_sample_idx += sampleToChunk.samples_per_chunk;
+    previous_sample_description_index = sampleToChunk.sample_description_index;
   }
 
   // --- read sample auxiliary information boxes
 
-  std::vector<std::shared_ptr<Box_saiz>> saiz_boxes = stbl->get_child_boxes<Box_saiz>();
-  std::vector<std::shared_ptr<Box_saio>> saio_boxes = stbl->get_child_boxes<Box_saio>();
+  std::vector<std::shared_ptr<Box_saiz> > saiz_boxes = stbl->get_child_boxes<Box_saiz>();
+  std::vector<std::shared_ptr<Box_saio> > saio_boxes = stbl->get_child_boxes<Box_saio>();
+
+  if (saio_boxes.size() != saiz_boxes.size()) {
+    return Error{
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Boxes 'saiz' and `saio` must come in pairs."
+    };
+  }
 
   for (const auto& saiz : saiz_boxes) {
     uint32_t aux_info_type = saiz->get_aux_info_type();
@@ -289,6 +409,14 @@ Track::Track(HeifContext* ctx, const std::shared_ptr<Box_trak>& trak_box)
     }
 
     if (saio) {
+      if (saio->get_num_samples() != saiz->get_num_samples()) {
+        return Error{
+          heif_error_Invalid_input,
+          heif_suberror_Unspecified,
+          "Number of samples in 'saiz' box does not match 'saio' box."
+        };
+      }
+
       if (aux_info_type == fourcc("suid")) {
         m_aux_reader_content_ids = std::make_unique<SampleAuxInfoReader>(saiz, saio);
       }
@@ -297,6 +425,13 @@ Track::Track(HeifContext* ctx, const std::shared_ptr<Box_trak>& trak_box)
         m_aux_reader_tai_timestamps = std::make_unique<SampleAuxInfoReader>(saiz, saio);
       }
     }
+    else {
+      return Error{
+        heif_error_Invalid_input,
+        heif_suberror_Unspecified,
+        "'saiz' box without matching 'saio' box."
+      };
+    }
   }
 
   // --- read track properties
@@ -314,21 +449,42 @@ Track::Track(HeifContext* ctx, const std::shared_ptr<Box_trak>& trak_box)
           heif_item_id id = box->get_item_ID();
 
           std::vector<uint8_t> data;
-          Error err = iloc->read_data(id, ctx->get_heif_file()->get_reader(), idat, &data, ctx->get_security_limits());
+          Error err = iloc->read_data(id, m_heif_context->get_heif_file()->get_reader(), idat, &data, m_heif_context->get_security_limits());
           if (err) {
             // TODO
           }
 
           Result contentIdResult = vector_to_string(data);
-          if (contentIdResult.error) {
+          if (!contentIdResult) {
             // TODO
           }
 
-          m_track_info.gimi_track_content_id = contentIdResult.value;
+          m_track_info.gimi_track_content_id = *contentIdResult;
         }
       }
     }
   }
+
+
+  // --- security checks
+
+  if (m_stsz->num_samples() > m_heif_context->get_security_limits()->max_sequence_frames) {
+    return {
+      heif_error_Memory_allocation_error,
+      heif_suberror_Security_limit_exceeded,
+      "Security limit for maximum number of sequence frames exceeded"
+    };
+  }
+
+
+  // --- initialize track tables
+
+  Error err = init_sample_timing_table();
+  if (err) {
+    return err;
+  }
+
+  return {};
 }
 
 
@@ -379,6 +535,15 @@ Track::Track(HeifContext* ctx, uint32_t track_id, const TrackOptions* options, u
   m_minf = std::make_shared<Box_minf>();
   mdia->append_child_box(m_minf);
 
+  // add (unused) 'dinf'
+
+  auto dinf = std::make_shared<Box_dinf>();
+  auto dref = std::make_shared<Box_dref>();
+  auto url = std::make_shared<Box_url>();
+  m_minf->append_child_box(dinf);
+  dinf->append_child_box(dref);
+  dref->append_child_box(url);
+
   // vmhd is added in Track_Visual
 
   m_stbl = std::make_shared<Box_stbl>();
@@ -390,6 +555,9 @@ Track::Track(HeifContext* ctx, uint32_t track_id, const TrackOptions* options, u
   m_stts = std::make_shared<Box_stts>();
   m_stbl->append_child_box(m_stts);
 
+  m_ctts = std::make_shared<Box_ctts>();
+  // The ctts box will be added in finalize_track(), but only there is frame-reordering.
+
   m_stsc = std::make_shared<Box_stsc>();
   m_stbl->append_child_box(m_stsc);
 
@@ -446,33 +614,106 @@ Track::Track(HeifContext* ctx, uint32_t track_id, const TrackOptions* options, u
 }
 
 
-std::shared_ptr<Track> Track::alloc_track(HeifContext* ctx, const std::shared_ptr<Box_trak>& trak)
+Result<std::shared_ptr<Track> > Track::alloc_track(HeifContext* ctx, const std::shared_ptr<Box_trak>& trak)
 {
   auto mdia = trak->get_child_box<Box_mdia>();
   if (!mdia) {
-    return nullptr;
+    return Error{
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Track has no 'mdia' box."
+    };
   }
 
   auto hdlr = mdia->get_child_box<Box_hdlr>();
-  if (!mdia) {
-    return nullptr;
+  if (!hdlr) {
+    return Error{
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "Track has no 'hdlr' box."
+    };
   }
 
+  std::shared_ptr<Track> track;
+
   switch (hdlr->get_handler_type()) {
     case fourcc("pict"):
     case fourcc("vide"):
-      return std::make_shared<Track_Visual>(ctx, trak);
+    case fourcc("auxv"):
+      track = std::make_shared<Track_Visual>(ctx);
+      break;
     case fourcc("meta"):
-      return std::make_shared<Track_Metadata>(ctx, trak);
-    default:
-      return nullptr;
+      track = std::make_shared<Track_Metadata>(ctx);
+      break;
+    default: {
+      std::stringstream sstr;
+      sstr << "Track with unsupported handler type '" << fourcc_to_string(hdlr->get_handler_type()) << "'.";
+      return Error{
+        heif_error_Unsupported_feature,
+        heif_suberror_Unsupported_track_type,
+        sstr.str()
+      };
+    }
   }
+
+  assert(track);
+  Error loadError = track->load(trak);
+  if (loadError) {
+    return loadError;
+  }
+
+  return {track};
 }
 
 
 bool Track::is_visual_track() const
 {
-  return m_handler_type == fourcc("pict");
+  return (m_handler_type == fourcc("pict") ||
+          m_handler_type == fourcc("vide"));
+}
+
+
+static const char* cAuxType_alpha_miaf = "urn:mpeg:mpegB:cicp:systems:auxiliary:alpha";
+static const char* cAuxType_alpha_hevc = "urn:mpeg:hevc:2015:auxid:1";
+static const char* cAuxType_alpha_avc = "urn:mpeg:avc:2015:auxid:1";
+
+heif_auxiliary_track_info_type Track::get_auxiliary_info_type() const
+{
+  if (m_auxiliary_info_type == cAuxType_alpha_miaf ||
+      m_auxiliary_info_type == cAuxType_alpha_hevc ||
+      m_auxiliary_info_type == cAuxType_alpha_avc) {
+    return heif_auxiliary_track_info_type_alpha;
+  }
+  else {
+    return heif_auxiliary_track_info_type_unknown;
+  }
+}
+
+const char* get_track_auxiliary_info_type(heif_compression_format format)
+{
+  switch (format) {
+    case heif_compression_HEVC:
+      return cAuxType_alpha_hevc;
+    case heif_compression_AVC:
+      return cAuxType_alpha_avc;
+    case heif_compression_AV1:
+      return cAuxType_alpha_miaf;
+    default:
+      return cAuxType_alpha_miaf; // TODO: is this correct for all remaining compression types ?
+  }
+}
+
+// TODO: is this correct or should we set the aux_info_type depending on the compression format?
+void Track::set_auxiliary_info_type(heif_auxiliary_track_info_type type)
+{
+  switch (type) {
+    case heif_auxiliary_track_info_type_alpha:
+      m_auxiliary_info_type = cAuxType_alpha_miaf;
+      break;
+    default:
+      m_auxiliary_info_type.clear();
+      break;
+  }
 }
 
 
@@ -489,24 +730,30 @@ uint32_t Track::get_first_cluster_sample_entry_type() const
 Result<std::string> Track::get_first_cluster_urim_uri() const
 {
   if (m_stsd->get_num_sample_entries() == 0) {
-    return Error{heif_error_Invalid_input,
-                 heif_suberror_Unspecified,
-                 "This track has no sample entries."};
+    return Error{
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "This track has no sample entries."
+    };
   }
 
   std::shared_ptr<const Box> sampleEntry = m_stsd->get_sample_entry(0);
   auto urim = std::dynamic_pointer_cast<const Box_URIMetaSampleEntry>(sampleEntry);
   if (!urim) {
-    return Error{heif_error_Usage_error,
-                 heif_suberror_Unspecified,
-                 "This cluster is no 'urim' sample entry."};
+    return Error{
+      heif_error_Usage_error,
+      heif_suberror_Unspecified,
+      "This cluster is no 'urim' sample entry."
+    };
   }
 
   std::shared_ptr<const Box_uri> uri = urim->get_child_box<const Box_uri>();
   if (!uri) {
-    return Error{heif_error_Invalid_input,
-                 heif_suberror_Unspecified,
-                 "The 'urim' box has no 'uri' child box."};
+    return Error{
+      heif_error_Invalid_input,
+      heif_suberror_Unspecified,
+      "The 'urim' box has no 'uri' child box."
+    };
   }
 
   return uri->get_uri();
@@ -515,17 +762,50 @@ Result<std::string> Track::get_first_cluster_urim_uri() const
 
 bool Track::end_of_sequence_reached() const
 {
-  return (m_next_sample_to_be_processed > m_chunks.back()->last_sample_number());
+  //return (m_next_sample_to_be_processed > m_chunks.back()->last_sample_number());
+  return m_next_sample_to_be_output >= m_num_output_samples;
 }
 
 
-void Track::finalize_track()
+Error Track::finalize_track()
 {
+  // --- write active chunk data
+
+  size_t data_start = m_heif_context->get_heif_file()->append_mdat_data(m_chunk_data);
+  m_chunk_data.clear();
+
+  // first sample in chunk? -> write chunk offset
+
+  if (true) {
+    // m_stsc->last_chunk_empty()) {
+    // TODO: we will have to call this at the end of a chunk to dump the current SAI queue
+
+    // if auxiliary data is interleaved, write it between the chunks
+    if (m_aux_helper_tai_timestamps) m_aux_helper_tai_timestamps->write_interleaved(get_file());
+    if (m_aux_helper_content_ids) m_aux_helper_content_ids->write_interleaved(get_file());
+
+    // TODO
+    assert(data_start < 0xFF000000); // add some headroom for header data
+    m_stco->add_chunk_offset(static_cast<uint32_t>(data_start));
+  }
+
+
+  // --- write rest of data
+
   if (m_aux_helper_tai_timestamps) m_aux_helper_tai_timestamps->write_all(m_stbl, get_file());
   if (m_aux_helper_content_ids) m_aux_helper_content_ids->write_all(m_stbl, get_file());
 
-  uint64_t duration = m_stts->get_total_duration(false);
+  uint64_t duration = m_stts->get_total_duration(true);
   m_mdhd->set_duration(duration);
+
+  m_stss->set_total_number_of_samples(m_stsz->num_samples());
+
+  // only add ctts box if we use frame-reordering
+  if (!m_ctts->is_constant_offset()) {
+    m_stbl->append_child_box(m_ctts);
+  }
+
+  return {};
 }
 
 
@@ -541,9 +821,34 @@ uint32_t Track::get_timescale() const
 }
 
 
-void Track::set_track_duration_in_movie_units(uint64_t total_duration)
+void Track::set_track_duration_in_movie_units(uint64_t total_duration, uint64_t segment_duration)
 {
   m_tkhd->set_duration(total_duration);
+
+  if (m_elst) {
+    Box_elst::Entry entry;
+    entry.segment_duration = segment_duration;
+
+    m_elst->add_entry(entry);
+  }
+}
+
+
+void Track::enable_edit_list_repeat_mode(bool enable)
+{
+  if (!m_elst) {
+    if (!enable) {
+      return;
+    }
+
+    auto edts = std::make_shared<Box_edts>();
+    m_trak->append_child_box(edts);
+
+    m_elst = std::make_shared<Box_elst>();
+    edts->append_child_box(m_elst);
+
+    m_elst->enable_repeat_mode(enable);
+  }
 }
 
 
@@ -570,38 +875,31 @@ void Track::set_sample_description_box(std::shared_ptr<Box> sample_description_b
 }
 
 
-Error Track::write_sample_data(const std::vector<uint8_t>& raw_data, uint32_t sample_duration, bool is_sync_sample,
-                               const heif_tai_timestamp_packet* tai, const std::string& gimi_contentID)
+Error Track::write_sample_data(const std::vector<uint8_t>& raw_data, uint32_t sample_duration,
+                               int32_t composition_time_offset,
+                               bool is_sync_sample,
+                               const heif_tai_timestamp_packet* tai, const std::optional<std::string>& gimi_contentID)
 {
-  size_t data_start = m_heif_context->get_heif_file()->append_mdat_data(raw_data);
-
-  // first sample in chunk? -> write chunk offset
-
-  if (m_stsc->last_chunk_empty()) {
-    // if auxiliary data is interleaved, write it between the chunks
-    if (m_aux_helper_tai_timestamps) m_aux_helper_tai_timestamps->write_interleaved(get_file());
-    if (m_aux_helper_content_ids) m_aux_helper_content_ids->write_interleaved(get_file());
-
-    // TODO
-    assert(data_start < 0xFF000000); // add some headroom for header data
-    m_stco->add_chunk_offset(static_cast<uint32_t>(data_start));
-  }
+  m_chunk_data.insert(m_chunk_data.end(), raw_data.begin(), raw_data.end());
 
   m_stsc->increase_samples_in_chunk(1);
 
-  m_stsz->append_sample_size((uint32_t)raw_data.size());
+  m_stsz->append_sample_size((uint32_t) raw_data.size());
 
   if (is_sync_sample) {
-    m_stss->add_sync_sample(m_next_sample_to_be_processed + 1);
+    m_stss->add_sync_sample(m_next_sample_to_be_output + 1);
   }
 
   if (sample_duration == 0) {
-    return {heif_error_Usage_error,
-            heif_suberror_Unspecified,
-            "Sample duration may not be 0"};
+    return {
+      heif_error_Usage_error,
+      heif_suberror_Unspecified,
+      "Sample duration may not be 0"
+    };
   }
 
   m_stts->append_sample_duration(sample_duration);
+  m_ctts->append_sample_offset(composition_time_offset);
 
 
   // --- sample timestamp
@@ -618,17 +916,19 @@ Error Track::write_sample_data(const std::vector<uint8_t>& raw_data, uint32_t sa
       m_aux_helper_tai_timestamps->add_nonpresent_sample();
     }
     else {
-      return {heif_error_Encoding_error,
-              heif_suberror_Unspecified,
-              "Mandatory TAI timestamp missing"};
+      return {
+        heif_error_Encoding_error,
+        heif_suberror_Unspecified,
+        "Mandatory TAI timestamp missing"
+      };
     }
   }
 
   // --- sample content id
 
   if (m_track_info.with_sample_content_ids != heif_sample_aux_info_presence_none) {
-    if (!gimi_contentID.empty()) {
-      auto id = gimi_contentID;
+    if (gimi_contentID) {
+      auto id = *gimi_contentID;
       const char* id_str = id.c_str();
       std::vector<uint8_t> id_vector;
       id_vector.insert(id_vector.begin(), id_str, id_str + id.length() + 1);
@@ -636,16 +936,20 @@ Error Track::write_sample_data(const std::vector<uint8_t>& raw_data, uint32_t sa
       if (err) {
         return err;
       }
-    } else if (m_track_info.with_sample_content_ids == heif_sample_aux_info_presence_optional) {
+    }
+    else if (m_track_info.with_sample_content_ids == heif_sample_aux_info_presence_optional) {
       m_aux_helper_content_ids->add_nonpresent_sample();
-    } else {
-      return {heif_error_Encoding_error,
-              heif_suberror_Unspecified,
-              "Mandatory ContentID missing"};
+    }
+    else {
+      return {
+        heif_error_Encoding_error,
+        heif_suberror_Unspecified,
+        "Mandatory ContentID missing"
+      };
     }
   }
 
-  m_next_sample_to_be_processed++;
+  m_next_sample_to_be_output++;
 
   return Error::Ok;
 }
@@ -662,77 +966,149 @@ void Track::add_reference_to_track(uint32_t referenceType, uint32_t to_track_id)
 }
 
 
-Result<heif_raw_sequence_sample*> Track::get_next_sample_raw_data()
+Error Track::init_sample_timing_table()
 {
-  if (m_current_chunk > m_chunks.size()) {
-    return Error{heif_error_End_of_sequence,
-                 heif_suberror_Unspecified,
-                 "End of sequence"};
-  }
+  m_num_samples = m_stsz->num_samples();
 
-  while (m_next_sample_to_be_processed > m_chunks[m_current_chunk]->last_sample_number()) {
-    m_current_chunk++;
+  // --- build media timeline
 
-    if (m_current_chunk > m_chunks.size()) {
-      return Error{heif_error_End_of_sequence,
-                   heif_suberror_Unspecified,
-                   "End of sequence"};
+  std::vector<SampleTiming> media_timeline;
+
+  uint64_t current_decoding_time = 0;
+  uint32_t current_chunk = 0;
+
+  for (uint32_t i = 0; i < m_num_samples; i++) {
+    SampleTiming timing;
+    timing.sampleIdx = i;
+    timing.media_decoding_time = current_decoding_time;
+    timing.sample_duration_media_time = m_stts->get_sample_duration(i);
+    current_decoding_time += timing.sample_duration_media_time;
+
+    while (current_chunk < m_chunks.size() &&
+           i > m_chunks[current_chunk]->last_sample_number()) {
+      current_chunk++;
+
+      if (current_chunk > m_chunks.size()) {
+        timing.chunkIdx = 0; // TODO: error
+      }
     }
+
+    timing.chunkIdx = current_chunk;
+
+    media_timeline.push_back(timing);
   }
 
-  const std::shared_ptr<Chunk>& chunk = m_chunks[m_current_chunk];
+  // --- build presentation timeline from editlist
 
-  DataExtent extent = chunk->get_data_extent_for_sample(m_next_sample_to_be_processed);
+  bool fallback = false;
+
+  if (m_heif_context->get_sequence_timescale() != get_timescale()) {
+    fallback = true;
+  }
+  else if (m_elst &&
+           m_elst->num_entries() == 1 &&
+           m_elst->get_entry(0).media_time == 0 &&
+           m_elst->get_entry(0).segment_duration == m_mdhd->get_duration() &&
+           m_elst->is_repeat_mode()) {
+    m_presentation_timeline = media_timeline;
+
+    uint64_t duration_media_units = get_duration_in_media_units();
+    if (duration_media_units == 0) {
+      return {
+        heif_error_Invalid_input,
+        heif_suberror_Unspecified,
+        "Track duration is zero."
+      };
+    }
+
+    m_num_output_samples = m_heif_context->get_sequence_duration() / get_duration_in_media_units() * media_timeline.size();
+  }
+  else {
+    fallback = true;
+  }
+
+  // Fallback: just play the media timeline
+  if (fallback) {
+    m_presentation_timeline = media_timeline;
+    m_num_output_samples = media_timeline.size();
+  }
+
+  return {};
+}
+
+
+Result<heif_raw_sequence_sample*> Track::get_next_sample_raw_data(const heif_decoding_options* options)
+{
+  uint64_t num_output_samples = m_num_output_samples;
+  if (options && options->ignore_sequence_editlist) {
+    num_output_samples = m_num_samples;
+  }
+
+  if (m_next_sample_to_be_output >= num_output_samples) {
+    return Error{
+      heif_error_End_of_sequence,
+      heif_suberror_Unspecified,
+      "End of sequence"
+    };
+  }
+
+  const auto& sampleTiming = m_presentation_timeline[m_next_sample_to_be_output % m_presentation_timeline.size()];
+  uint32_t sample_idx = sampleTiming.sampleIdx;
+  uint32_t chunk_idx = sampleTiming.chunkIdx;
+
+  const std::shared_ptr<Chunk>& chunk = m_chunks[chunk_idx];
+
+  DataExtent extent = chunk->get_data_extent_for_sample(sample_idx);
   auto readResult = extent.read_data();
-  if (readResult.error) {
-    return readResult.error;
+  if (!readResult) {
+    return readResult.error();
   }
 
   heif_raw_sequence_sample* sample = new heif_raw_sequence_sample();
-  sample->data = *readResult.value;
+  sample->data = **readResult;
 
   // read sample duration
 
   if (m_stts) {
-    sample->duration = m_stts->get_sample_duration(m_next_sample_to_be_processed);
+    sample->duration = m_stts->get_sample_duration(sample_idx);
   }
 
   // --- read sample auxiliary data
 
   if (m_aux_reader_content_ids) {
-    auto readResult = m_aux_reader_content_ids->get_sample_info(get_file().get(), m_next_sample_to_be_processed);
-    if (readResult.error) {
-      return readResult.error;
+    auto readResult = m_aux_reader_content_ids->get_sample_info(get_file().get(), sample_idx);
+    if (!readResult) {
+      return readResult.error();
     }
 
-    if (!readResult.value.empty()) {
-      Result<std::string> convResult = vector_to_string(readResult.value);
-      if (convResult.error) {
-        return convResult.error;
+    if (!readResult->empty()) {
+      Result<std::string> convResult = vector_to_string(*readResult);
+      if (!convResult) {
+        return convResult.error();
       }
 
-      sample->gimi_sample_content_id = convResult.value;
+      sample->gimi_sample_content_id = *convResult;
     }
   }
 
   if (m_aux_reader_tai_timestamps) {
-    auto readResult = m_aux_reader_tai_timestamps->get_sample_info(get_file().get(), m_next_sample_to_be_processed);
-    if (readResult.error) {
-      return readResult.error;
+    auto readResult = m_aux_reader_tai_timestamps->get_sample_info(get_file().get(), sample_idx);
+    if (!readResult) {
+      return readResult.error();
     }
 
-    if (!readResult.value.empty()) {
-      auto resultTai = Box_itai::decode_tai_from_vector(readResult.value);
-      if (resultTai.error) {
-        return resultTai.error;
+    if (!readResult->empty()) {
+      auto resultTai = Box_itai::decode_tai_from_vector(*readResult);
+      if (!resultTai) {
+        return resultTai.error();
       }
 
       sample->timestamp = heif_tai_timestamp_packet_alloc();
-      heif_tai_timestamp_packet_copy(sample->timestamp, &resultTai.value);
+      heif_tai_timestamp_packet_copy(sample->timestamp, &*resultTai);
     }
   }
 
-  m_next_sample_to_be_processed++;
+  m_next_sample_to_be_output++;
 
   return sample;
 }
